// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:flutter/material.dart';
import 'package:wyatt_ui_components/src/domain/entities/entities.dart';
part 'component_theme_data.g.dart';

/// {@template component_theme_data}
/// A class that holds all the components that are used in the app.
/// {@endtemplate}
@CopyWith()
class ComponentThemeData {
  /// {@macro component_theme_data}
  factory ComponentThemeData({
    TopAppBarComponent? topAppBar,
    TopNavigationBarComponent? topNavigationBar,
    BottomNavigationBarComponent? bottomNavigationBar,
    ErrorComponent? error,
    LoaderComponent? loader,
    RichTextBuilderComponent? richTextBuilder,
    TextInputComponent? textInput,
    FileSelectionButtonComponent? fileSelectionButton,
    FlatButtonComponent? flatButton,
    SimpleIconButtonComponent? simpleIconButton,
    SymbolButtonComponent? symbolButton,
    InformationCardComponent? informationCard,
    PortfolioCardComponent? portfolioCard,
    QuoteCardComponent? quoteCard,
    SkillCardComponent? skillCard,
  }) =>
      ComponentThemeData.raw(
        topAppBar: topAppBar,
        topNavigationBar: topNavigationBar,
        bottomNavigationBar: bottomNavigationBar,
        error: error,
        loader: loader,
        richTextBuilder: richTextBuilder,
        textInput: textInput,
        fileSelectionButton: fileSelectionButton,
        flatButton: flatButton,
        simpleIconButton: simpleIconButton,
        symbolButton: symbolButton,
        informationCard: informationCard,
        portfolioCard: portfolioCard,
        quoteCard: quoteCard,
        skillCard: skillCard,
      );

  /// {@macro component_theme_data}
  factory ComponentThemeData.fromOther(ComponentThemeData other) =>
      ComponentThemeData(
        topAppBar: other.topAppBar,
        topNavigationBar: other.topNavigationBar,
        bottomNavigationBar: other.bottomNavigationBar,
        error: other.error,
        loader: other.loader,
        richTextBuilder: other.richTextBuilder,
        textInput: other.textInput,
        fileSelectionButton: other.fileSelectionButton,
        flatButton: other.flatButton,
        simpleIconButton: other.simpleIconButton,
        symbolButton: other.symbolButton,
        informationCard: other.informationCard,
        portfolioCard: other.portfolioCard,
        quoteCard: other.quoteCard,
        skillCard: other.skillCard,
      );

  /// Create a [ComponentThemeData] given a set of exact values. Most values
  /// must be specified.
  ///
  /// This will rarely be used directly. It is used by lerp to
  /// create intermediate themes based on two themes created with the
  /// [ComponentThemeData] constructor.
  const ComponentThemeData.raw({
    this.topAppBar,
    this.topNavigationBar,
    this.bottomNavigationBar,
    this.error,
    this.loader,
    this.richTextBuilder,
    this.textInput,
    this.fileSelectionButton,
    this.flatButton,
    this.simpleIconButton,
    this.symbolButton,
    this.informationCard,
    this.portfolioCard,
    this.quoteCard,
    this.skillCard,
  });

  R _get<T extends Component, R>(T? component, R? returned) {
    if (component == null) {
      throw FlutterError('No $T component provided.\n'
          'Please provide a $T component in the ComponentThemeData.');
    }

    if (returned == null) {
      throw FlutterError('$T does not have a CopyWith method.\n'
          'Please provide a $T component that implements CopyWith '
          'in the ComponentThemeData.');
    }

    return returned;
  }

  // Bars
  final TopAppBarComponent? topAppBar;
  $TopAppBarComponentCWProxy get topAppBarComponent =>
      _get(topAppBar, topAppBar?.copyWith);
  final TopNavigationBarComponent? topNavigationBar;
  $TopNavigationBarComponentCWProxy get topNavigationBarComponent =>
      _get(topNavigationBar, topNavigationBar?.copyWith);
  final BottomNavigationBarComponent? bottomNavigationBar;
  $BottomNavigationBarComponentCWProxy get bottomNavigationBarComponent =>
      _get(bottomNavigationBar, bottomNavigationBar?.copyWith);

  // CRUD Widgets
  final ErrorComponent? error;
  $ErrorComponentCWProxy get errorComponent => _get(error, error?.copyWith);
  final LoaderComponent? loader;
  $LoaderComponentCWProxy get loaderComponent => _get(loader, loader?.copyWith);

  // Cards
  final InformationCardComponent? informationCard;
  $InformationCardComponentCWProxy get informationCardComponent =>
      _get(informationCard, informationCard?.copyWith);
  final PortfolioCardComponent? portfolioCard;
  $PortfolioCardComponentCWProxy get portfolioCardComponent =>
      _get(portfolioCard, portfolioCard?.copyWith);
  final QuoteCardComponent? quoteCard;
  $QuoteCardComponentCWProxy get quoteCardComponent =>
      _get(quoteCard, quoteCard?.copyWith);
  final SkillCardComponent? skillCard;
  $SkillCardComponentCWProxy get skillCardComponent =>
      _get(skillCard, skillCard?.copyWith);

  // Rich Text
  final RichTextBuilderComponent? richTextBuilder;
  $RichTextBuilderComponentCWProxy get richTextBuilderComponent =>
      _get(richTextBuilder, richTextBuilder?.copyWith);

  // Text Inputs
  final TextInputComponent? textInput;
  $TextInputComponentCWProxy get textInputComponent =>
      _get(textInput, textInput?.copyWith);

  // Buttons
  final FileSelectionButtonComponent? fileSelectionButton;
  $FileSelectionButtonComponentCWProxy get fileSelectionButtonComponent =>
      _get(fileSelectionButton, fileSelectionButton?.copyWith);
  final FlatButtonComponent? flatButton;
  $FlatButtonComponentCWProxy get flatButtonComponent =>
      _get(flatButton, flatButton?.copyWith);
  final SimpleIconButtonComponent? simpleIconButton;
  $SimpleIconButtonComponentCWProxy get simpleIconButtonComponent =>
      _get(simpleIconButton, simpleIconButton?.copyWith);
  final SymbolButtonComponent? symbolButton;
  $SymbolButtonComponentCWProxy get symbolButtonComponent =>
      _get(symbolButton, symbolButton?.copyWith);
}
